(deffunction ask-question (?question $?allowed-values)
                    (printout t ?question)
                    (bind ?answer (read))
                    (if (lexemep ?answer)
                        then
                            (bind ?answer (lowcase ?answer)))
                            (while (not (member ?answer ?allowed-values)) do

                                   (printout t ?question)
                                   (bind ?answer (read))
                                   (if (lexemep ?answer)
                                       then
                                           (bind ?answer (lowcase ?answer))))
                                           ?answer
                                          )
	
  (deffunction yes-or-no-p (?question)
                    (bind ?response (ask-question ?question yes no y n))
                    (if (or (eq ?response yes) (eq ?response y))
                        then
                            TRUE
                        else
                            FALSE)
                    )
	
(defrule determine-engine-state ""
                (not (working-state engine ?))
                (not (repair ?))
                =>
                 (if (yes-or-no-p "    (yes/no)? ")
                   then                           
                       (assert (working))
                   else
                       (assert (working-state engine normal)))
       )

(defrule determine-state ""
                (not (working ?))
                (not (repair ?))
                =>
                   (if (yes-or-no-p "   (yes/no)? ")
                   then
                       (assert (working-state engine normal))
                   
                   else
                       (assert (working-state engine unsatisfactory))
                   else
                       (assert (working-state engine does-not-start)))
                   
          )
	

(defrule determine-rotation-state ""                
                (working-state engine does-not-start)                
                (not (rotation-state engine ?))
                (not (repair ?))
                =>
                 (if (yes-or-no-p "   (yes/no)? ")
                     then
		         (assert (rolation-state engine rotates))
                         
                         (assert (spark-state engine irregular-spark))
                     else
                        
                         (assert (rotation-state engine does-not-rotate))
		  
                         (assert (spark-state engine does-not-spark)))
             )
	
(defrule determine-gas-level ""
                 (working-state engine does-not-start)
                 (rotation-state engine rotates)            
                                
                (not (repair ?))
                =>
                (if (not (yes-or-no-p "      (yes/no)? "))
                then
       
                    (assert (repair "  .")))           
            )

	
(defrule determine-battery-state ""
                (rotation-state engine does-not-rotate)
                (not (charge-state battery ?))
                (not (repair ?))
                =>
                 (if (yes-or-no-p ",     (yes/no)? ")
                 then
        
                     (assert (charge-state battery charged))
                 else	          
                     (assert (repair " ."))
                            
	             (assert (charge-state battery dead)))
             )
	
(defrule determine-low-output ""
                (working-state engine unsatisfactory)                
	    (not (symptom engine low-output | not-low-output))
                (not (repair ?))
                =>
                 (if (yes-or-no-p "      (yes/no)? ")
                     then
	              
                         (assert (symptom engine low-output))
                     else
	              
                         (assert (symptom engine not-low-output)))
         )
	
(defrule determine-point-surface-state ""
                (or (and (working-state engine does-not-start) 
                         (spark-state engine irregular-spark)) 
                    (symptom engine not-low-output)) 
                (not (repair ?))
                =>
                 (bind ?response (ask-question "      
                                      (1  /2  /3 )?"

                                     1 2 3))
                 (if (eq ?response 2)
                     then
	             
                         (assert (repair "  -  ."))
                     else
                         (if (eq ?response 3)
                     then
		 
                         (assert (repair "  -  ."))))

        )
	
(defrule determine-conductivity-test ""
                (working-state engine does-not-start)
                (spark-state engine does-not-spark)
                (charge-state battery charged) 
                (not (repair ?))
                =>
                 (if (yes-or-no-p "     (yes/no)? ")
                 then
	         
                     (assert (repair "  ."))
                 else
	          

                     (assert (repair "  .")))
         )
	
(defrule determine-sluggishness ""
                (working-state engine unsatisfactory)
                (not (repair ?))
                =>

                 (if (yes-or-no-p "     
                     (      (yes/no)? ")
                     then
 		 
                         (assert (repair "    .")))
        )
	
(defrule determine-misfiring ""
                (working-state engine unsatisfactory)

                (not (repair ?))
                =>
                 (if (yes-or-no-p "     (yes/no)? ")
                     then
		 
                         (assert (repair "   ."))
                         (assert (spark-state engine irregular-spark))) 
        )
	
(defrule determine-knocking ""
                (working-state engine unsatisfactory)
                (not (repair ?))
                =>
                 (if (yes-or-no-p ",     (yes/no)? ")
                     then
		 
                         (assert (repair "  .")))
                      
        )


(defrule normal-engine-state-conclusions ""

              (declare (salience 10))
	        
	      (working-state engine normal)
                =>
                 (assert (repair "   .")) 
                 (assert (spark-state engine normal)) 
                 (assert (charge-state battery charged)) 
                 (assert (rolation-state engine rotates)) 
        )

(defrule unsatisfactory-engine-state-conclusions ""
                (declare (salience 10))
	    
                (working-state engine unsatisfactory)
                =>
                 (assert (charge-state battery charged)) 
                 (assert (rotation-state engine rotates)) 
        )
	
(defrule no-repairs ""
            (declare (salience -10))
            (not (repair ?))
            =>
            (assert (repair "   .   :
  
             .
  ,        ,   .
    ,     .
       . 
           .
         
           .
           , 
             .
              .
          , 
        . 
           
        .
          
               
        !

   .     !"))
        )
	
(defrule print-repair ""
                (declare (salience 10))
                (repair ?item)
                =>
                 (printout t crlf crlf)
                 (printout t " :")
                 (printout t crlf crlf)
                 (format t " %s%n%n%n" ?item)
         )
	
(deffacts init
              (troubleshoot-mode engine)
              (menu-level engine main)
         )
	
(defrule main-menu ""
            (declare (salience 500))
            (troubleshoot-mode engine)
            ?ml <- (menu-level engine main)
            =>
                (retract ?ml)
                (printout t crlf crlf )
                (printout t "                                   " crlf)
                (printout t "                                                                   " crlf)
                (printout t "                                1)  " crlf)

                (printout t "                                2)   " crlf) 
                (printout t "                                                                    " crlf)
                (printout t " : " crlf)
                (bind ?response (read))
  (assert (problem-response engine ?response))
  (printout t crlf crlf )
          )
	
(defrule user-exit ""
            (troubleshoot-mode engine)

            (problem-response engine 2)
            =>
               (printout t " !" crlf)
               (halt)
          )
	
(defrule begin-start ""
            (troubleshoot-mode engine)
            ?pr <- (problem-response engine 1)
            =>
                (retract ?pr)
                (assert (defrule determine-engine-state))

         )
	
(defrule system-banner ""
                (declare (salience 10))
                =>
                 (printout t crlf crlf)
                 (printout t "                                *************************************" crlf)
                 (printout t "                                *    *" crlf)
                 (printout t "                                *************************************" crlf)

                 (printout t "                                                                                            " crlf)
                 (printout t "                 " crlf) 
                 (printout t "                  ." crlf)
                 (printout t "               YES  NO.                                         " crlf)
                 (printout t crlf crlf)
        )