(deffunction ask-question (?question $?allowed-values) (printout t ?question) (bind ?answer (read))
               (if (lexemep ?answer)
                then
               (bind ?answer (lowcase ?answer)))
               (while (not (member ?answer ?allowed-values)) do
               (printout t ?question) (bind ?answer (read))
               (if (lexemep ?answer)
                then
               (bind ?answer (lowcase ?answer)))) ?answer)

(deffunction yes-or-no-p (?question)
              (bind ?response (ask-question ?question yes no y n))
              (if (or (eq ?response yes) (eq ?response y))
              then
              TRUE
              else
              FALSE))

(deffacts initial-state
  (square 1 1 " " corner)
  (square 2 1 " " side)
  (square 3 1 " " corner)
  (square 1 2 " " side)
  (square 2 2 " " center)
  (square 3 2 " " side)
  (square 1 3 " " corner)
  (square 2 3 " " side)
  (square 3 3 " " corner))

(defrule evaluate-potential-win ""
  (evaluate board for ?)
  (square ?row1 ?col1 ?pl&~" " ?)
  (square ?row2 ?col2 " " ?)
  (test (or  (!= ?row1 ?row2) (!= ?col1 ?col2)))
  (square ?row3 ?col3 " " ?type)
  (test (and (or  (!= ?row1 ?row3) (!= ?col1 ?col3))
	            (or  (!= ?row2 ?row3) (!= ?col2 ?col3))))
  (test (or  (and (= ?row1 ?row2)
		                (= ?row1 ?row3))
	            (and (= ?col1 ?col2)
		                (= ?col1 ?col3))
	            (and (= (+ ?row1 ?col1) (+ ?row2 ?col2))
		                (= (+ ?row1 ?col1) (+ ?row3 ?col3)))
	            (and (= (- ?row1 ?col1) (- ?row2 ?col2))
		                (= (- ?row1 ?col1) (- ?row3 ?col3)))))
   =>
   (assert (potential-win ?row3 ?col3 ?type ?pl)))

(defrule evaluate-immediate-block-or-win ""
  (evaluate board for ?)
  (square ?row1 ?col1 ?pl&~" " ?)
  (square ?row2 ?col2 ?pl ?)
  (test (or (!= ?row1 ?row2) (!= ?col1 ?col2)))
  (square ?row3 ?col3 " " ?type)
  (test (and (or  (!= ?row1 ?row3) (!= ?col1 ?col3))
	            (or  (!= ?row2 ?row3) (!= ?col2 ?col3))))
  (test (or  (and (= ?row1 ?row2)
		                (= ?row1 ?row3))
	            (and (= ?col1 ?col2)
		                (= ?col1 ?col3))
	            (and (= (+ ?row1 ?col1) (+ ?row2 ?col2))
		                (= (+ ?row1 ?col1) (+ ?row3 ?col3)))
	            (and (= (- ?row1 ?col1) (- ?row2 ?col2))
		                (= (- ?row1 ?col1) (- ?row3 ?col3)))))
   =>
   (assert (immediate-block-or-win ?row3 ?col3 ?type ?pl)))

(defrule tie-game ""
  ?phase <- (evaluate board for ?)
  (not (square ? ? " " ?))
  =>
  (retract ?phase)
  (printout t "!" t)
 (if (yes-or-no-p "   (yes/no)? ")
                      then
                     (reset)
                     (run)
                     else
                     (printout t " !     !" crlf) 
                     (halt)))

(defrule move-computer-to-win
   ?phase <- (player-move computer)
   (computer is ?cs)
   ?dummy <- (immediate-block-or-win ?row ?col ?type ?cs)
   ?square <- (square ?row ?col ? ?)
   =>
   (assert (display board))
   (retract ?dummy ?phase ?square)
   (assert (square ?row ?col ?cs ?type))
   (printout t "  !" t)
  (if (yes-or-no-p "   (yes/no)? ")
                      then
                     (reset)
                     (run)
                     else
                     (printout t " !     !" crlf) 
                     (halt)))

(defrule move-computer-to-block ""
  ?phase <- (player-move computer)
  (computer is ?cs)
  (not (immediate-block-or-win ? ? ? ?cs))
  (human is ?hs)
  ?dummy <- (immediate-block-or-win ?row ?col ?type ?hs)
  ?square <- (square ?row ?col ? ?)
  =>
  (retract ?dummy ?phase ?square)
  (assert (display board))
  (assert (square ?row ?col ?cs ?type))
  (assert (next-move human)))

(defrule move-to-center ""
   ?phase <- (player-move computer)
   (computer is ?cs)
   (not (immediate-block-or-win ? ? ? ?))
   ?square <- (square ? ? " " center)
   =>
   (retract ?phase ?square)
   (assert (display board))
   (assert (square 2 2 ?cs center))
   (assert (next-move human)))

(defrule move-to-corner ""
   (computer is ?cs)
   ?phase <- (player-move computer)
   (not (immediate-block-or-win ? ? ? ?)) 
   ?square <- (square ?row ?col " " corner)
   (not (square ? ? " " center))
   =>
   (retract ?phase ?square)
   (assert (display board))
   (assert (square ?row ?col ?cs corner))
   (assert (next-move human)))

(defrule move-to-side ""
   ?phase <- (player-move computer)
   (computer is ?cs)
   (not (immediate-block-or-win ? ? ? ?)) 
   ?square <- (square ?row ?col " " side) 
   (not (square ? ? " " corner))
   (not (square ? ? " " center))
   =>
   (retract ?phase ?square)
   (assert (display board))
   (assert (square ?row ?col ?cs side))
   (assert (next-move human)))

(defrule human-input ""
  (player-move human)
  =>
  (printout t " ? ")
  (bind ?col (read))
  (printout t " ? ")
  (bind ?row (read))
  (printout t crlf)
  (assert (human-to ?col ?row)))

(defrule valid-human-move ""
  ?phase <- (player-move human)
  (human is ?hs)
  ?move  <- (human-to ?row ?col)
  ?square <- (square ?row ?col " " ?type)
  =>
  (assert (display board))
  (retract ?phase ?move ?square)
  (assert (square ?row ?col ?hs ?type))
  (assert (valid-move ?row ?col)))

(defrule human-winning-move ""
  ?move <- (valid-move ?row ?col)
  (human is ?hs)
  ?info <- (immediate-block-or-win ?row ?col ?type ?hs)
  =>
  (retract ?move ?info)
  (printout t "!  !   !" t)
 (if (yes-or-no-p "   (yes/no)? ")
                      then
                     (reset)
                     (run)
                     else
                     (printout t " !     !" crlf) 
                     (halt)))

(defrule human-blocking-move ""
  ?move <- (valid-move ?row ?col)
  (computer is ?cs)
  ?info <- (immediate-block-or-win ? ? ?type ?cs)
  =>
  (retract ?info ?move)
  (assert (next-move computer)))

(defrule no-block-or-win-for-human ""
  ?move <- (valid-move ?row ?col)
  (not (immediate-block-or-win ? ? ? ?))
  =>
  (retract ?move)
  (assert (next-move computer)))

(defrule invalid-human-move ""
  ?phase <- (player-move human)
  ?move  <- (human-to ?row ?col)
  (square ?row ?col ~" " ?)
  =>
  (printout t "  ." t)
  (retract ?phase ?move)
  (assert (player-move human)))

(defrule cleanup1 ""
  (cleanup for ?)
  ?trash <- (immediate-block-or-win ? ? ? ?)
  =>
  (retract ?trash))

(defrule cleanup2 ""
  (cleanup for ?)
  ?trash <- (potential-win ? ? ? ?)
  =>
  (retract ?trash))

(defrule setup ""
  =>
  (printout t "  ? ")
  (bind ?resp (read))
  (printout t t)
  (if (or (eq ?resp yes) (eq ?resp y))
      then
      (assert (computer is o))
      (assert (human is x))
      (assert (evaluate board for computer))
      else
      (assert (computer is x))
      (assert (human is o))
      (assert (evaluate board for human))))

(defrule cleanup-for-player ""
 (declare (salience -10))
  ?phase <- (next-move ?player)
  =>
  (retract ?phase)
  (assert (cleanup for ?player)))

(defrule switch-to-evaluate ""
 (declare (salience -10))
  ?phase <- (cleanup for ?player)
  =>
  (retract ?phase)
  (assert (evaluate board for ?player)))

(defrule switch-to-move ""
  (evaluate board for computer)
 ?phase <- (evaluate board for ?player)
  =>
  (retract ?phase)
  (printout t "*****************" t)
  (printout t "   " t)
  (printout t "*****************" t)
  (assert (player-move ?player)))

(defrule switch-to-move2 ""
  (evaluate board for human)
  ?phase <- (evaluate board for ?player)
  =>
  (retract ?phase)
  (printout t "*****************" t)
  (printout t "   " t)
  (printout t "*****************" t)
  (assert (player-move ?player)))

(defrule board ""
  ?phase <- (display board)
  (square 1 1 ?p1-1 ?)
  (square 2 1 ?p2-1 ?)
  (square 3 1 ?p3-1 ?)
  (square 1 2 ?p1-2 ?)
  (square 2 2 ?p2-2 ?)
  (square 3 2 ?p3-2 ?)
  (square 1 3 ?p1-3 ?)
  (square 2 3 ?p2-3 ?)
  (square 3 3 ?p3-3 ?)
  =>
  (retract ?phase)
  (format t "%n :%n%n")
  (format t " %s|%s|%s%n" ?p1-1 ?p2-1 ?p3-1)
  (format t " -----%n")
  (format t " %s|%s|%s%n" ?p1-2 ?p2-2 ?p3-2)
  (format t " -----%n")
  (format t " %s|%s|%s%n%n" ?p1-3 ?p2-3 ?p3-3))
