(deffunction yes-or-no-p (?question)
  (bind ?x bogus)
  (while (and (neq ?x yes) (neq ?x y) (neq ?x no) (neq ?x n))
     (format t "%s((y)  (n)) " ?question)
     (bind ?x (lowcase (sym-cat (read)))))
  (if (or (eq ?x yes) (eq ?x y)) then TRUE else FALSE))

(deftemplate win-totals
  (slot human (type INTEGER) (default 0))
  (slot computer (type INTEGER) (default 0))
  (slot ties (type INTEGER) (default 0)))

(deftemplate results
   (slot winner (type SYMBOL) (allowed-symbols rock paper scissors))
   (slot loser (type SYMBOL) (allowed-symbols rock paper scissors))
   (slot why (type STRING)))

(deffacts information
  (results (winner rock) (loser scissors) (why "   "))
  (results (winner scissors) (loser paper) (why "  "))
  (results (winner paper) (loser rock) (why "  "))
  (valid-answer rock r rock)
  (valid-answer paper p paper)
  (valid-answer scissors s scissors))

(defrule startup
  =>
  (printout t "                          , ,  ?!" crlf crlf)
  (printout t "  ,     : ,   !" crlf)
  (printout t "   !" crlf)
  (printout t "  !" crlf)
  (printout t "  !" crlf crlf)
  (set-strategy random)
  (assert (win-totals))
  (assert (get-human-move)))

(defrule get-human-move
  (get-human-move)
  =>
  (printout t "           " crlf)
  (printout t "  :  (s),  (p)   (r) ? ")
  (assert (human-choice (read))))

(defrule good-human-move
  ?f1 <- (human-choice ?choice)
  (valid-answer ?answer $? =(lowcase ?choice) $?)
  ?f2 <- (get-human-move)
  =>
  (retract ?f1 ?f2)
  (assert (human-choice ?answer))
  (assert (get-computer-move)))

(defrule bad-human-move
  ?f1 <- (human-choice ?choice)
  (not (valid-answer ?answer $? =(lowcase ?choice) $?))
  ?f2 <- (get-human-move)
  =>
  (retract ?f1 ?f2)
  (assert (get-human-move)))

(defrule computer-picks-rock
   ?f1 <- (get-computer-move)
   =>
   (printout t "           " crlf)
   (printout t "  - " crlf)
   (retract ?f1)
   (assert (computer-choice rock))
   (assert (determine-results)))

(defrule computer-picks-paper
   ?f1 <- (get-computer-move)
   =>
   (printout t "           " crlf)
   (printout t "  - " crlf)
   (retract ?f1)
   (assert (computer-choice paper))
   (assert (determine-results)))

(defrule computer-picks-scissors
   ?f1 <- (get-computer-move)
   =>
   (printout t "           " crlf)
   (printout t "  - " crlf)
   (retract ?f1)
   (assert (computer-choice scissors))
   (assert (determine-results)))

(defrule computer-wins
  ?f1 <- (determine-results)
  ?f2 <- (computer-choice ?cc)
  ?f3 <- (human-choice ?hc)
  ?w <- (win-totals (computer ?cw))
  (results (winner ?cc) (loser ?hc) (why ?explanation))
  =>
  (retract ?f1 ?f2 ?f3)
  (modify ?w (computer (+ ?cw 1)))
  (format t "%s%n" ?explanation)
  (printout t "           " crlf)
  (printout t ",  ,    !" t)
  (assert (determine-play-again)))

(defrule human-wins
  ?f1 <- (determine-results)
  ?f2 <- (computer-choice ?cc)
  ?f3 <- (human-choice ?hc)
  ?w <- (win-totals (human ?hw))
  (results (winner ?hc) (loser ?cc) (why ?explanation))
  =>
  (retract ?f1 ?f2 ?f3)
  (modify ?w (human (+ ?hw 1)))
  (format t "%s%n" ?explanation)
  (printout t "           " crlf)
  (printout t " . ,  !" t)
  (assert (determine-play-again)))

(defrule tie
  ?f1 <- (determine-results)
  ?f2 <- (computer-choice ?cc)
  ?f3 <- (human-choice ?cc)
  ?w <- (win-totals (ties ?nt))
  =>
  (retract ?f1 ?f2 ?f3)
  (modify ?w (ties (+ ?nt 1)))
  (printout t "           " crlf)
  (printout t "!" t)
  (assert (determine-play-again)))

(defrule play-again
  ?f1 <- (determine-play-again)
  (win-totals (computer ?ct) (human ?ht) (ties ?tt))
  =>
  (retract ?f1)
  (assert (get-human-move))
  (printout t "           " crlf)
  (if (not (yes-or-no-p " ? ")) 
     then 
     (printout t "      " crlf)
     (printout t "---------------------------------------------- " crlf)
     (printout t "                    " crlf)
     (printout t "-----------------------------------------------           " crlf)
     (printout t "  " ?ht " ()." t)
     (printout t "  " ?ct " ()." t)
     (printout t " " ?tt " ()."  t)
     (printout t "-----------------------------------------------           " crlf)
     (printout t "  :" (+ ?ht ?ct ?tt ) " ()."  t)
     (printout t "-----------------------------------------------           " crlf)
     (halt)))
